package gus05.entity.gus.file.editor.classe.editor2;

import java.io.*;
import java.util.*;

import javax.swing.JTextArea;


    /*
     * magic            4 This identifies the .class file format. It should be 0xCAFEBABE. If its anything else, youre dealing with a format more recent than this book.
     * minor version    2 The minor version of the compiler
     * major version    2 The major version of the compiler
     * constant pool    variable
     * 
     * The first two bytes give the number of entries in the constant pool.
     * Then, as many bytes as are necessary to fill that many entries are read.
     * The constant pool is a table of constant values used by this class.
     * 
     * access flags     2 These bit flags tell you whether the class is public, final, abstract, an interface, and a few other things.
     * this class       2 This tells you which entry in the constant pool holds this classs class info.
     * superclass       2 If this is zero, then this classs only superclass is java.lang.Object. Otherwise, this is an index into the constant pool for the superclass class info.
     * interfaces       variable
     * 
     * The interface table holds two byte indices into the constant pool table, one for each interface that this class implements.
     * The first two bytes give the number of entries in the interface table. Therefore, after reading the first two bytes,
     * you have to read twice as many bytes as the number stored in the first two bytes.
     * 
     * fields           variable
     * 
     * The fields table includes one fields info structure for each field in the class.
     * 
     * methods          variable
     * 
     * The method table contains the byte codes for each method in the class,
     * the return type of the method, and the types of each argument to the method.
     * 
     * attributes       2 The attributes of the class
     */

public class ClassFileParser {

    public static final int JAVA_MAGIC = 0xCAFEBABE;
    public static final char CLASS_DESCRIPTOR = 'L';
    
    public static final int CONSTANT_UTF8 = 1;
    public static final int CONSTANT_UNICODE = 2;
    public static final int CONSTANT_INTEGER = 3;
    public static final int CONSTANT_FLOAT = 4;
    public static final int CONSTANT_LONG = 5;
    public static final int CONSTANT_DOUBLE = 6;
    public static final int CONSTANT_CLASS = 7;
    public static final int CONSTANT_STRING = 8;
    public static final int CONSTANT_FIELD = 9;
    public static final int CONSTANT_METHOD = 10;
    public static final int CONSTANT_INTERFACEMETHOD = 11;
    public static final int CONSTANT_NAMEANDTYPE = 12;
    
    public static final int ACCESS_PUBLIC = 0x0001;
    public static final int ACCESS_PRIVATE = 0x0002;
    public static final int ACCESS_PROTECTED = 0x0004;
    public static final int ACCESS_STATIC = 0x0008;
    public static final int ACCESS_FINAL = 0x0010;
    public static final int ACCESS_SYNCHRONIZED = 0x0020;
    public static final int ACCESS_VOLATILE = 0x0040;
    public static final int ACCESS_TRANSIENT = 0x0080;
    public static final int ACCESS_NATIVE = 0x0100;
    public static final int ACCESS_ABSTRACT = 0x0400;
    
    
    
    
    private File classFile;
    private HashSet<String> dependencies;
    
    private int minorVersion;
    private int majorVersion;
    
    private Constant[] constantPool;
    
    private Constant thisClass;
    private Constant superClass;
    private Constant[] interfaces;
    
    private String className;
    private String superClassName;
    private String interfaceNames[];
    
    private Element[] fields;
    private Element[] methods;
    private Attribute[] attributes;
    
    
    
    
    private DataInputStream in;
    
    
    private byte _byte() throws IOException
    {return in.readByte();}
    
    private int _unsigned() throws IOException
    {return in.readUnsignedShort();}
    
    private short _short() throws IOException
    {return in.readShort();}

    private int _int() throws IOException
    {return in.readInt();}
    
    private float _float() throws IOException
    {return in.readFloat();}
    
    private long _long() throws IOException
    {return in.readLong();}
    
    private double _double() throws IOException
    {return in.readDouble();}
    
    private String _UTF8() throws IOException
    {return in.readUTF();}
    
    
    
    
    
    
    
    private JTextArea area;
    
    private void println(String m)
    {area.append(m+"\n");}
    
    private void delim()
    {println("--------------------------------------");}
    
    public ClassFileParser(JTextArea area)
    {
        this.area = area;
        dependencies = new HashSet<String>();
    }
    
    
    
    
    
    
    
    public void parse(File classFile) throws Exception
    {
        this.classFile = classFile;
        area.setText("");
        dependencies.clear();
        
        FileInputStream is = new FileInputStream(classFile);
        in = new DataInputStream(is);

        delim();
        //parse java magic
        int magic = _int();
        if(magic!=JAVA_MAGIC) throw new Exception("Invalid magic number for class file: "+classFile);
        println("magic number = "+magic);
        
        // parse minor version
        minorVersion = _unsigned();
        //if(minorVersion!=3) throw new Exception("Invalid minor version for class file: "+classFile);
        println("minor version = "+minorVersion);
        
        // parse major version
        majorVersion = _unsigned();
        // if(majorVersion!=45) throw new Exception("Invalid major version for class file: "+classFile);
        println("major version = "+majorVersion);
        
        // parse constant pool
        parseConstantPool();
        
        // parse access flags
        parseAccessFlags();
        
        // parse class name
        parseClassName();
        
        // parse superclass name
        parseSuperClassName();
        
        // parse interface names
        parseInterfaces();
        
        parseFields();
        parseMethods();
        parseAttributes();
        addClassConstantReferences();

        in.close();
    }

    
    
    
  
    
    
    
    
    private void parseConstantPool() throws Exception
    {
        int nb = _unsigned();
        delim();
        println("constant pool size = "+nb);
        
        constantPool = new Constant[nb];
        for (int i=1;i<nb;i++)
        {
            constantPool[i] = new Constant();
            // 8-byte constants use two constant pool entries
            byte tag = constantPool[i].tag;
            if(tag==CONSTANT_DOUBLE || tag==CONSTANT_LONG) i++;
        }
    }
    
    
    

    private void parseAccessFlags() throws Exception
    {
        int accessFlags = _unsigned();
        delim();
        println("access flags = "+accessFlags);
        
        boolean isPublic =           ((accessFlags & 0x0001)!=0);
        boolean isFinal =            ((accessFlags & 0x0010)!=0);
        boolean isSpecial =          ((accessFlags & 0x0020)!=0);
        boolean isInterface =        ((accessFlags & 0x0200)!=0);
        boolean isAbstract =         ((accessFlags & 0x0400)!=0);
        
        if (isInterface && !isAbstract)     throw new Exception("This interface is not abstract");
        if (isAbstract && isFinal)          throw new Exception("This class is abstract and final");
        
        if(isPublic)        println("is public");
        if(isFinal)         println("is final");
        if(isInterface)     println("is interface");
        if(isAbstract)      println("is abstract");
        if(isSpecial)       println("is special");
    }

    
    
    
    
    
    
    private void parseClassName() throws IOException
    {
        int index = _unsigned();
        thisClass = constantPool[index];
        className = getClassConstantName(thisClass);
        
        delim();
        println("index = "+index);
        println("class name = "+className);
    }
    
    
    
    
    /*
     * if this class is the java.lang.Object (the only class without a superclass)
     * then the index into the constant pool will be zero.
     */
    
    private void parseSuperClassName() throws IOException
    {
        int index = _unsigned();
        superClass = constantPool[index];
        superClassName = getClassConstantName(superClass);
        
        delim();
        println("super class name = "+superClassName);
        addDependency(superClassName);
    }
    
    
    
    
    

    private void parseInterfaces() throws IOException
    {
        int n = _unsigned();
        delim();
        println("implements "+n+" interfaces");
        
        interfaces = new Constant[n];
        interfaceNames = new String[n];
        
        for (int i=0;i<n;i++)
        {
            int index = _unsigned();
            println("index = "+index);
            interfaces[i] = constantPool[index];
            interfaceNames[i] = getClassConstantName(interfaces[i]);
            
            println("interface name = "+interfaceNames[i]);
            addDependency(interfaceNames[i]);
        }
    }

    
    
    
    
    
    
    private void parseFields() throws IOException
    {
        delim();
        int n = _unsigned();
        println("FIELDS: "+n+" fields");
        fields = new Element[n];
        
        for(int i=0;i<n;i++)
        {
            fields[i] = new Element();

            String[] types = descriptorToTypes(fields[i].descriptor);
            for(int t=0;t<types.length;t++) addDependency(types[t]);
        }
    }
    
    
    
    
    
    
    

    private void parseMethods() throws IOException
    {
        delim();
        int n = _unsigned();
        println("METHODS: "+n+" methods");
        methods = new Element[n];
        
        for (int i=0;i<n;i++)
        {
            methods[i] = new Element();

            String[] types = descriptorToTypes(methods[i].descriptor);
            for (int t=0;t<types.length;t++)
            if(types[t].length()>0) addDependency(types[t]);
        }
    }
    
    
    
    
    
    
    
    
    

    private void parseAttributes() throws IOException
    {
        int attributesCount = _unsigned();
        attributes = new Attribute[attributesCount];

        for(int i=0;i<attributesCount;i++)
        {
            attributes[i] = new Attribute();
            // Section 4.7.7 of VM Spec - Class File Format
            if(attributes[i].name!=null)
                if(attributes[i].name.equals("SourceFile"))
                {
                    byte[] b = attributes[i].value;
                    int b0 = b[0] < 0 ? b[0] + 256 : b[0];
                    int b1 = b[1] < 0 ? b[1] + 256 : b[1];
                    int pe = b0 * 256 + b1;

                    String descriptor = toUTF8(pe);
                    println("source file = "+descriptor);
                }
        }
    } 
    
    
    
    
    
    
    
    
    
    

    private void addClassConstantReferences() throws IOException
    {
        for(int j=1;j<constantPool.length;j++)
        {
            int tag = constantPool[j].tag;
            if(tag==CONSTANT_CLASS)
            {
                String name = toUTF8(constantPool[j].nameIndex);
                addDependency(name);
            }
            if(tag==CONSTANT_DOUBLE || tag==CONSTANT_LONG) j++;
        }
    }

    
    
    
    
    
    
    
    
    private String getClassConstantName(Constant entry) throws IOException
    {
        if(entry==null)return "";
        return toUTF8(entry.nameIndex);
    }
    
    
    
    
    
    
    
    

    private String toUTF8(int index) throws IOException
    {
        Constant entry = constantPool[index];
        if(entry.tag==CONSTANT_UTF8) return (String)entry.value;
        throw new IOException("Constant pool entry is not a UTF8 type: "+index);
    }
    
    
    
    
    
    
    
    
    

    private void addDependency(String dependingClass)
    {
        //if(dependingClass==null) return;
        dependingClass = dependingClass.replace('/', '.');
        println("DEPENDENCY: "+dependingClass);
        dependencies.add(dependingClass);
    }
    
    
    
    

   
    
    
    
    
    
    
    
    

    private String[] descriptorToTypes(String descriptor)
    {
        int typesCount = 0;
        for(int index=0;index<descriptor.length();index++)
        if(descriptor.charAt(index)==';') typesCount++;

        String types[] = new String[typesCount];

        int typeIndex = 0;
        for(int index=0;index<descriptor.length();index++)
        {
            int startIndex = descriptor.indexOf(CLASS_DESCRIPTOR,index);
            if(startIndex<0) break;
            
            index = descriptor.indexOf(';',startIndex+1);
            types[typeIndex++] = descriptor.substring(startIndex+1,index);
        }
        return types;
    }
    
    
    
    
    
    
    

    
    
    
    
    
    
    
    
    

    private class Element
    {
        public int accessFlags;
        public int nameIndex;
        public int descriptorIndex;
        public String descriptor;
        
        public Attribute[] attr;

        public Element() throws IOException
        {
            accessFlags = _unsigned();
            nameIndex = _unsigned();
            descriptorIndex = _unsigned();
            descriptor = toUTF8(descriptorIndex);
            
            println("Element "+descriptor+": access="+accessFlags+" nameIndex="+nameIndex+" descriptorIndex="+descriptorIndex);
            
            int n = _unsigned();
            attr = new Attribute[n];
            for(int i=0;i<n;i++) attr[i] = new Attribute();
        }
    }

    
    
    
    
    
    private class Attribute
    {
        public String name;
        public byte[] value;
        
        public Attribute() throws IOException
        {
            int nameIndex = _unsigned();
            if(nameIndex!=-1) name = toUTF8(nameIndex);

            int length = _int();
            value = new byte[length];
            for(int i=0;i<length;i++)
            value[i] = _byte();
        }
    }
    
    
    
    
    
    
    private class Constant
    {
        public byte tag;
        public String description;
        public int nameIndex = -1;
        public int typeIndex = -1;
        public Object value = null;
        
        public Constant() throws Exception
        {
            tag = _byte();
            switch (tag)
            {
            case (CONSTANT_CLASS):              init("class",_unsigned());break;
            case (CONSTANT_STRING):             init("string",_unsigned());break;
            case (CONSTANT_FIELD):              init("field",_unsigned(),_unsigned());break;
            case (CONSTANT_METHOD):             init("method",_unsigned(),_unsigned());break;
            case (CONSTANT_INTERFACEMETHOD):    init("interface method",_unsigned(),_unsigned());break;
            case (CONSTANT_NAMEANDTYPE):        init("name and type",_unsigned(),_unsigned());break;
            case (CONSTANT_INTEGER):            init("integer",new Integer(_int()));break;
            case (CONSTANT_FLOAT):              init("float",new Float(_float()));break;
            case (CONSTANT_LONG):               init("long",new Long(_long()));break;
            case (CONSTANT_DOUBLE):             init("double",new Double(_double()));break;
            case (CONSTANT_UTF8):               init("UTF8",_UTF8());break;

            default: throw new Exception("Unknown constant: "+tag);
            }
        }
        
        private void init(String _description, int _nameIndex)
        {
            description = _description;
            nameIndex = _nameIndex;
            println("const: "+description+" (tag="+tag+") nameIndex="+nameIndex);
        }

        private void init(String _description, Object _value)
        {
            description = _description;
            value = _value;
            println("const: "+description+" (tag="+tag+") value="+value);
        }

        private void init(String _description, int _nameIndex, int _typeIndex)
        {
            description = _description;
            nameIndex = _nameIndex;
            typeIndex = _typeIndex;
            println("const: "+description+" (tag="+tag+") nameIndex="+nameIndex+" typeIndex="+typeIndex);
        }

        public String toString()
        {
            StringBuffer s = new StringBuffer(description+"\n");
            s.append("tag: "+tag+"\n");
            if(nameIndex>-1) s.append(" nameIndex: "+nameIndex+"\n");
            if(typeIndex>-1) s.append(" typeIndex: "+typeIndex+"\n");
            if(value!=null)  s.append(" value: "+value+"\n");
            return s.toString();
        }
    }
    
    
    
    
    
    
    

    
    public String toString()
    {
        StringBuffer s = new StringBuffer();

        try
        {
            s.append("\n" + className + ":\n");
            s.append("\nConstants:\n");
            for(int i=1;i<constantPool.length;i++)
            {
                Constant entry = constantPool[i];
                s.append("    "+i+". "+entry.toString()+"\n");
                if (entry.tag==CONSTANT_DOUBLE || entry.tag==CONSTANT_LONG) i++;
            }

            s.append("\nClass Name: "+className+"\n");
            s.append("Super Name: "+superClassName+"\n\n");
            
            s.append(interfaceNames.length + " interfaces\n");
            for(int i=0;i<interfaceNames.length;i++)
            s.append("    " + interfaceNames[i]+"\n");
            
            s.append("\n"+fields.length+" fields\n");
            for(int i=0;i<fields.length;i++)
            s.append(fields[i].toString()+"\n");
            
            s.append("\n" + methods.length + " methods\n");
            for(int i=0;i<methods.length;i++)
            s.append(methods[i].toString()+"\n");
            
//            s.append("\nDependencies:\n");
//            Iterator imports = jClass.getImportedPackages().iterator();
//            while(imports.hasNext())
//            {
//                JavaPackage jPackage = (JavaPackage) imports.next();
//                s.append("    "+jPackage.getName()+"\n");
//            }
        }
        catch(Exception e)
        {e.printStackTrace();}
        
        return s.toString();
    }
    
}
